# -*- coding: utf-8 -*-

"""
koordinates.client
==================
"""

import copy
import json
import logging
import os
import re
import sys

import pkg_resources
import requests
import requests_toolbelt
import six

from . import layers, licenses, publishing, sets, tokens, users, catalog, sources, exports
from . import exceptions


logger = logging.getLogger(__name__)


class Client(object):
    """
    A `Client` is used to define the host and api-version which the user
    wants to connect to. The user identity is also defined when `Client`
    is instantiated.
    """
    def __init__(self, host, token=None, activate_logging=False):
        '''
        :param str host: the domain name of the Koordinates site to connect to (eg. ``labs.koordinates.com``)
        :param str token: Koordinates API token to use for authentication
        :param bool activate_logging: if True then logging to stderr is activated
        '''
        if activate_logging:
            logging.basicConfig(stream=sys.stderr,
                                level=logging.DEBUG,
                                format='%(asctime)s %(levelname)s %(module)s %(message)s')

        logger.debug('Initializing Client object for %s', host)

        self.host = host

        if token:
            self.token = token
        elif 'KOORDINATES_TOKEN' in os.environ:
            self.token = os.environ['KOORDINATES_TOKEN']
        else:
            raise KeyError('No authentication token specified, and KOORDINATES_TOKEN not available in the environment.')

        self._user_agent = requests_toolbelt.user_agent('KoordinatesPython', pkg_resources.require("koordinates")[0].version)

        self._init_managers(public={
                'sets': sets.SetManager,
                'publishing': publishing.PublishManager,
                'tokens': tokens.TokenManager,
                'layers': layers.LayerManager,
                'tables': layers.TableManager,
                'licenses': licenses.LicenseManager,
                'catalog': catalog.CatalogManager,
                'sources': sources.SourceManager,
                'exports': exports.ExportManager,
            },
            private=(
                users.GroupManager,
                users.UserManager,
                sources.ScanManager,
                sources.DatasourceManager,
                exports.CropFeatureManager,
                exports.CropLayerManager,
            )
        )

        self._session = requests.Session()
        self._session.headers.update({
            'Accept': 'application/json',
            'User-Agent': self._user_agent,
        })
        if self.token:
            self._session.headers['Authorization'] = 'key {token}'.format(token=self.token)

        super(self.__class__, self).__init__()

    def _init_managers(self, public, private):
        self._manager_map = {}
        for manager_class in private:
            mgr = manager_class(self)
            self._register_manager(mgr.model, mgr)

        for alias, manager_class in public.items():
            mgr = manager_class(self)
            self._register_manager(mgr.model, mgr)
            setattr(self, alias, mgr)

    def _register_manager(self, model, manager):
        self._manager_map[model] = manager

    def get_manager(self, model):
        """
        Return the active manager for the given model.
        :param model: Model class to look up the manager instance for.
        :return: Manager instance for the model associated with this client.
        """
        if isinstance(model, six.string_types):
            # undocumented string lookup
            for k, m in self._manager_map.items():
                if k.__name__ == model:
                    return m
            else:
                raise KeyError(model)

        return self._manager_map[model]


    def _assemble_headers(self, method, user_headers=None):
        """
        Takes the supplied headers and adds in any which
        are defined at a client level and then returns
        the result.

        :param user_headers: a `dict` containing headers defined at the
                             request level, optional.

        :return: a `dict` instance
        """

        headers = copy.deepcopy(user_headers or {})

        if method not in ('GET', 'HEAD'):
            headers.setdefault('Content-Type', 'application/json')

        return headers

    def request(self, method, url, *args, **kwargs):
        headers = self._assemble_headers(method, kwargs.pop("headers", {}))
        r = self._raw_request(method, url, headers, *args, **kwargs)

        if method == 'POST':
            # If we're posting to an endpoint
            if r.status_code in (requests.codes.created, requests.codes.accepted):
                # and we get a 201/202 response
                if 'location' in r.headers:
                    # with a location header
                    logger.info("%s -> %s", r.status_code, r.headers['location'])

                    # follow it and return the results of the GET
                    r2 = self.request('GET', r.headers['location'])
                    r2.parent_response = r
                    return r2

        return r

    def _raw_request(self, method, url, headers, *args, **kwargs):
        # for the Koordinates library logging, strip auth tokens from log messages
        # and log POST/PUT bodies if we're sending JSON.
        # Get low-level logging via the requests.packages.urllib3 logger.
        if 'json' in kwargs:
            logger.info('Request: %s %s headers=%s body=%s', method, url, json.dumps(headers), json.dumps(kwargs['json']))
        else:
            logger.info('Request: %s %s headers=%s', method, url, json.dumps(headers))

        try:
            r = self._session.request(method, url, headers=headers, *args, **kwargs)
            logger.info('Response: %d %s in %s', r.status_code, r.reason, r.elapsed)
            logger.debug('Response: headers=%s', r.headers)
            r.raise_for_status()
            return r
        except requests.HTTPError as e:
            logger.warn('Response: %s: %s', e, r.text)
            raise exceptions.ServerError.from_requests_error(e)
        except requests.RequestException as e:
            raise exceptions.ServerError.from_requests_error(e)

    def get_url_path(self, datatype, verb, urltype, params={}, api_version=None):
        api_version = api_version or 'v1'
        templates = getattr(self, 'URL_TEMPLATES__%s' % api_version)

        url = templates[datatype][verb][urltype]
        return url.format(**params)

    def reverse_url(self, datatype, url, verb='GET', urltype='single', api_version=None):
        """
        Extracts parameters from a populated URL

        :param datatype: a string identifying the data the url accesses.
        :param url: the fully-qualified URL to extract parameters from.
        :param verb: the HTTP verb needed for use with the url.
        :param urltype: an adjective used to the nature of the request.
        :return: dict
        """
        api_version = api_version or 'v1'
        templates = getattr(self, 'URL_TEMPLATES__%s' % api_version)

        # this is fairly simplistic, if necessary we could use the parse lib
        template_url = r"https://(?P<api_host>.+)/services/api/(?P<api_version>.+)"
        template_url += re.sub(r'{([^}]+)}', r'(?P<\1>.+)', templates[datatype][verb][urltype])
        # /foo/{foo_id}/bar/{id}/
        m = re.match(template_url, url or '')
        if not m:
            raise KeyError("No reverse match from '%s' to %s.%s.%s" % (url, datatype, verb, urltype))

        r = m.groupdict()
        del r['api_host']
        if r.pop('api_version') != api_version:
            raise ValueError("API version mismatch")
        return r

    def get_url(self, datatype, verb, urltype, params={}, api_host=None, api_version=None):
        """Returns a fully formed url

        :param datatype: a string identifying the data the url will access.
        :param verb: the HTTP verb needed for use with the url.
        :param urltype: an adjective used to the nature of the request.
        :param \*\*params: substitution variables for the URL.
        :return: string
        :rtype: A fully formed url.
        """
        api_version = api_version or 'v1'
        api_host = api_host or self.host

        subst = params.copy()
        subst['api_host'] = api_host
        subst['api_version'] = api_version

        url = "https://{api_host}/services/api/{api_version}"
        url += self.get_url_path(datatype, verb, urltype, params, api_version)
        return url.format(**subst)

    URL_TEMPLATES__v1 = {
        'LAYER': {
            'GET': {
                'singleversion': '/layers/{layer_id}/versions/{version_id}/',
                'single': '/layers/{id}/',
                'multi': '/layers/',
                'multidraft': '/layers/drafts/',
            },
            'POST': {
                'create': '/layers/',
                'update': '/layers/{layer_id}/versions/import/',
            },
            'DELETE': {
                'delete': '/layers/{id}/',
            }
        },
        'SET': {
            'GET': {
                'single': '/sets/{id}/',
                'metadata': '/sets/{id}/metadata',
                'multi': '/sets/',
            },
            'POST': {
                'create': '/sets/',
            },
            'PUT': {
                'update': '/sets/{id}/',
            },
        },
        'VERSION': {
            'GET': {
                'single': '/layers/{layer_id}/versions/{version_id}/',
                'multi': '/layers/{layer_id}/versions/',
                'draft': '/layers/{layer_id}/versions/draft/',
                'published': '/layers/{layer_id}/versions/published/',
            },
            'POST': {
                'create': '/layers/{layer_id}/versions/',
                'import': '/layers/{layer_id}/versions/{version_id}/import/',
                'publish': '/layers/{layer_id}/versions/{version_id}/publish/',
            },
            'PUT': {
                'edit': '/layers/{layer_id}/versions/{version_id}/',
            },
            'DELETE': {
                'single': '/layers/{layer_id}/versions/{version_id}/',
            },
        },
        'CATALOG': {
            'GET': {
                'multi': '/data/',
                'latest': '/data/latest/',
            },
        },
        'PUBLISH': {
            'GET': {
                'single': '/publish/{id}/',
                'multi': '/publish/',
            },
            'POST': {
                'create': '/publish/',
            },
            'DELETE': {
                'single': '/publish/{id}/',
            }
        },
        'TOKEN': {
            'GET': {
                'single': '/tokens/{id}/',
                'multi': '/tokens/',
            },
            'POST': {
                'create': '/tokens/',
            },
            'PUT': {
                'update': '/tokens/{id}/',
            },
            'DELETE': {
                'single': '/tokens/{id}/',
            },
        },
        'LICENSE': {
            'GET': {
                'single': '/licenses/{id}/',
                'multi': '/licenses/',
                'cc': '/licenses/{slug}/{jurisdiction}/',
            },
        },
        'METADATA': {
            # InnerManager, so relative to a parent object
            'GET': {
                'get': 'metadata/',
            },
            'POST': {
                'set': 'metadata/',
            }
        },
        'SOURCE': {
            'GET': {
                'single': '/sources/{id}/',
                'multi': '/sources/',
            },
            'POST': {
                'create': '/sources/',
            },
            'DELETE': {
                'single': '/sources/{id}/',
            }
        },
        'SCAN': {
            'GET': {
                'single': '/sources/{source_id}/scans/{scan_id}/',
                'multi': '/sources/{source_id}/scans/',
                'log': '/sources/{source_id}/scans/{scan_id}/log/',
                'all': '/sources/scans/',
            },
            'DELETE': {
                'cancel': '/sources/{source_id}/scans/{scan_id}/',
            },
            'POST': {
                'create': '/sources/{source_id}/scans/',
            }
        },
        'DATASOURCE': {
            'GET': {
                'multi': '/sources/{source_id}/datasources/',
                'single': '/sources/{source_id}/datasources/{datasource_id}/',
            }
        },
        # InnerManager, so relative to a parent object
        'PERMISSION': {
            'GET': {
                'multi': 'permissions/',
                'single': 'permissions/{permission_id}/',
            },
            'POST': {
                'single': 'permissions/',
            },
            'PUT': {
                'multi': 'permissions/',
            }
        },
        'EXPORT': {
            'GET': {
                'multi': '/exports/',
                'single': '/exports/{id}/',
            },
            'POST': {
                'create': '/exports/',
                'validate': '/exports/validate/',
            },
            'OPTIONS': {
                'options': '/exports/',
            },
            'DELETE': {
                'single': '/exports/{id}/',
            }
        },
        'CROPLAYER': {
            'GET': {
                'multi': '/exports/croplayers/',
                'single': '/exports/croplayers/{id}/',
            },
        },
        'CROPFEATURE': {
            'GET': {
                'multi': '/exports/croplayers/{croplayer_id}/cropfeatures/',
                'single': '/exports/croplayers/{croplayer_id}/cropfeatures/{cropfeature_id}/',
            },
        },
    }



# -*- coding: utf-8 -*-

"""
koordinates.sets
================

The `Sets API <https://help.koordinates.com/api/publisher-admin-api/sets-api/>`_
is used for grouping layers, tables and documents together.
"""

import logging

from koordinates.permissions import PermissionObjectMixin
from koordinates.users import Group
from koordinates.metadata import Metadata, MetadataManager
from koordinates import base
from koordinates.utils import is_bound


logger = logging.getLogger(__name__)


class SetManager(base.Manager):
    """
    Accessor for querying Sets.

    Access via the ``sets`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = 'SET'

    def __init__(self, client):
        super(SetManager, self).__init__(client)
        # Inner model managers
        self._metadata = MetadataManager(self, client)

    def create(self, set):
        """
        Creates a new Set.
        """
        target_url = self.client.get_url('SET', 'POST', 'create')
        r = self.client.request('POST', target_url, json=set._serialize())
        return set._deserialize(r.json(), self)

    def set_metadata(self, set_id, fp):
        """
        Set the XML metadata on a set.

        :param file fp: file-like object to read the XML metadata from.
        """
        base_url = self.client.get_url('SET', 'GET', 'single', {'id': set_id})
        self._metadata.set(base_url, fp)


class Set(base.Model, PermissionObjectMixin):
    '''
    Represents a single set grouping of layers, tables, and documents.
    '''
    class Meta:
        manager = SetManager
        serialize_skip = ('permissions',)
        deserialize_skip = ('permissions',)

    def _deserialize(self, data, manager):
        super(Set, self)._deserialize(data, manager)
        self.group = Group()._deserialize(data['group'], manager.client.get_manager(Group)) if data.get("group") else None
        self.metadata = Metadata()._deserialize(data["metadata"], manager._metadata, self) if data.get("metadata") else None
        return self

    @is_bound
    def save(self):
        """
        Saves changes to a Set.
        """
        target_url = self._client.get_url('SET', 'PUT', 'update', {'id': self.id})
        r = self._client.request('PUT', target_url, json=self._serialize())
        return self._deserialize(r.json(), self._manager)

    @is_bound
    def set_metadata(self, fp):
        """
        Set the XML metadata on a set.

        :param file fp: file-like object to read the XML metadata from.
        """
        base_url = self._client.get_url('SET', 'GET', 'single', {'id': self.id})
        self._manager._metadata.set(base_url, fp)

        # reload myself
        r = self._client.request('GET', base_url)
        return self._deserialize(r.json(), self._manager)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import re
from codecs import open


version = ''
with open('koordinates/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)
with open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()

setup(
    name='koordinates',
    packages=['koordinates',],
    version=version,
    description='A Python client library for a number of Koordinates web APIs',
    long_description=readme,
    author='Koordinates Limited',
    author_email='support@koordinates.com',
    url='https://github.com/koordinates/python-client',
    keywords='koordinates api',
    license = 'BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: GIS',
    ],
    entry_points = {
        'console_scripts': [
            'koordinates-create-token = koordinates.tokens:console_create',
        ],
    },
    install_requires=[
        'python-dateutil>=2,<3',
        'pytz',
        'requests>=2.5,<3',
        'requests-toolbelt>=0.4,<1',
        'six>=1.10.0,<2',
    ],
    extras_require={
        ':python_version=="2.7"': [
            'contextlib2',
        ],
    },
    tests_require=[
        'pytest>=3.3',
        'responses>=0.3',
        'coverage>=3.7,<4',
    ],
    zip_safe=False,
)




# -*- coding: utf-8 -*-

"""
koordinates.token
=================

Support for creating and managing API tokens via the `Token API`__.

__ https://help.koordinates.com/api/publisher-admin-api/token-api/

"""
from __future__ import absolute_import

import logging

from koordinates import base
from koordinates.utils import make_date
from .utils import is_bound

logger = logging.getLogger(__name__)


class TokenManager(base.Manager):
    """
    Accessor for querying Tokens.

    Access via the ``tokens`` property of a :py:class:`koordinates.client.Client` instance.
    """


    _URL_KEY = 'TOKEN'

    def delete(self, id ):
        """ Delete a token """
        target_url = self.client.get_url('TOKEN', 'DELETE', 'single', {'id':id})

        r = self.client.request('DELETE', target_url, headers={'Content-type': 'application/json'})
        r.raise_for_status()


    def create(self, token, email, password):
        """
        Create a new token

        :param Token token: Token instance to create.
        :param str email: Email address of the Koordinates user account.
        :param str password: Koordinates user account password.
        """
        target_url = self.client.get_url('TOKEN', 'POST', 'create')
        post_data = {
            'grant_type': 'password',
            'username': email,
            'password': password,
            'name': token.name,
        }
        if getattr(token, 'scope', None):
            post_data['scope'] = token.scope
        if getattr(token, 'expires_at', None):
            post_data['expires_at'] = token.expires_at
        if getattr(token, 'referrers', None):
            post_data['referrers'] = token.referrers

        r = self.client._raw_request('POST', target_url, json=post_data, headers={'Content-type': 'application/json'})
        return self.create_from_result(r.json())


class Token(base.Model):
    '''
    Represents an API Token.
    '''
    class Meta:
        manager = TokenManager
        serialize_skip = ('key',)

    @property
    def scopes(self):
        """ Read/Write accessor for the :py:attr:`.scope` property as a list of scope strings. """
        return (self.scope or '').split()

    @scopes.setter
    def scopes(self, value):
        self.scope = ' '.join(value or [])

    def __str__(self):
        s = "%s (%s): %s" % (self.id, self.name, self.key_hint)
        return s

    @is_bound
    def save(self):
        """ Saves changes to a token """
        target_url = self._client.get_url('TOKEN', 'PUT', 'update', {'id': self.id})
        r = self._client.request('PUT', target_url, json=self._serialize())
        return self._deserialize(r.json(), self._manager)


def console_create():
    """
    Command line tool (``koordinates-create-token``) used to create an API token.
    """
    import argparse
    import getpass
    import re
    import sys
    import requests
    from six.moves import input
    from koordinates.client import Client

    parser = argparse.ArgumentParser(description="Command line tool to create a Koordinates API Token.")
    parser.add_argument('site', help="Domain (eg. labs.koordinates.com) for the Koordinates site.", metavar='SITE')
    parser.add_argument('email', help="User account email address", metavar='EMAIL')
    parser.add_argument('name', help="Description for the key", metavar='NAME')
    parser.add_argument('--scopes', help="Scopes for the new API token", nargs='+', metavar="SCOPE")
    parser.add_argument('--referrers', help="Restrict the request referrers for the token. You can use * as a wildcard, eg. *.example.com", nargs='+', metavar='HOST')
    parser.add_argument('--expires', help="Expiry time in ISO 8601 (YYYY-MM-DD) format", metavar="DATE")
    args = parser.parse_args()

    # check we have a valid-ish looking domain name
    if not re.match(r'(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}$)', args.site):
        parser.error("'%s' doesn't look like a valid domain name." % args.site)

    # check we have a valid-ish looking email address
    if not re.match(r'[^@]+@[^@]+\.[^@]+$', args.email):
        parser.error("'%s' doesn't look like a valid email address." % args.email)

    password = getpass.getpass('Account password for %s: ' % args.email)
    if not password:
        parser.error("Empty password.")

    expires_at = make_date(args.expires)

    print("\nNew API Token Parameters:")
    print("  Koordinates Site: %s" % args.site)
    print("  User email address: %s" % args.email)
    print("  Name: %s" % args.name)
    print("  Scopes: %s" % ' '.join(args.scopes or ['(default)']))
    print("  Referrer restrictions: %s" % ' '.join(args.referrers or ['(none)']))
    print("  Expires: %s" % (expires_at or '(never)'))
    if input("Continue? [Yn] ").lower() == 'n':
        sys.exit(1)

    token = Token(name=args.name)
    if args.scopes:
        token.scopes = args.scopes
    if args.referrers:
        token.referrers = args.referrers
    if expires_at:
        token.expires_at = expires_at

    print("\nRequesting token...")
    # need a dummy token here for initialisation
    client = Client(host=args.site, token='-dummy-')
    try:
        token = client.tokens.create(token, args.email, password)
    except requests.HTTPError as e:
        print("%s: %s" % (e, e.response.text))

        # Helpful tips for specific errors
        if e.response.status_code == 401:
            print("  => Check your email address, password, and site domain carefully.")

        sys.exit(1)

    print("Token created successfully.\n  Key: %s\n  Scopes: %s" % (token.key, token.scope))
